// Stack.java
//
// CompSci 141 / CSE 141 / Informatics 101 Fall 2009
// In-class example: a generic stack class in Java
//
// Stack<E> is a generic implementation of a stack in Java.  There are two
// interesting things going on in this class:
//
//   (1) It is generic.  When you create a new stack object, you can tell
//       the compiler specifically what kind of stack you're interested in
//       by supplying a type parameter, e.g., Stack<String>.  (The type
//       parameter must be a class type, though you can use "wrapper classes"
//       such as Integer to store primitive values, e.g. Stack<Integer>.
//       Java is fairly adept at converting ints to Integers and back again
//       using a feature called "autoboxing.")  Why we care about generics is
//       because they allow us to better express the intent behind our programs.
//       When we intend a stack to contain nothing but strings, we can specify
//       this by creating a Stack<String>.  This arms readers of our program
//       (including ourselves) with knowledge that helps them to understand it.
//       More importantly, it also arms the compiler with information that can
//       be used to generate compile-time errors when we do things in our
//       program that violate our stated intentions, such as pushing an integer
//       to a Stack<String>.  The compiler, in general, promises to enforce our
//       intent, so that it will never be possible to put anything but a String
//       into a Stack<String>; this promise drives many of the seemingly
//       esoteric rules about generics in Java.
//
//   (2) It supports iteration in the same way that the various classes in the
//       Java Collections Framework do.  There are two benefits to this.  From
//       a design perspective, this is a big win because it will make the
//       class easier to use for Java programmers who are already familiar with
//       the typical interfaces in the Java Collections Framework.  Using
//       established idioms when they are appropriate is a great way to make
//       your code more understandable.  On the practical side, supporting
//       iteration in this way allows this class to plug into existing language
//       features such as the "foreach" loop and library methods that take
//       standard Java Iterators as parameters.  (We could take this a step
//       further by having our Stack<E> class extend from one of the collection
//       interfaces/classes, such as AbstractCollection<E>, but iteration is
//       enough for this example.)


import java.util.*;


// Notice that I've declared that the Stack<E> class implements an interface
// called Iterable<E>.  This is a standard interface in Java that is to be
// implemented by classes whose objects can be iterated using a standard
// Java Iterator.  Naturally, this is true of most "collection" classes
// (i.e., data structures that store a collection of objects).  A stack
// certainly qualifies.
//
// Implementing the Iterable<E> interface requires us to include the following
// method in our class:
//
//     public Iterator<E> iterator()
//
// which creates and returns a new iterator over this stack.  Our iterator will
// return the top element of the stack first and the bottom element last.
//
// This stack is implemented using an ArrayList, with the bottom item appearing
// in cell 0 of the list and the top item appearing in cell n-1 (if there are
// n items in the stack).

public class Stack<E> implements Iterable<E>
{
	// Within the Stack class, "E" refers to the type parameter used to
	// instantiate a particular Stack object.  For a Stack<String>, you can
	// think of "E" as meaning "String"; for a Stack<ActivationRecord>, you
	// can think of "E" as meaning "ActivationRecord."


	// ArrayLists are generic, just like our Stack is.  To express our intent,
	// we should make it clear that the ArrayList should store the appropriate
	// kind of element for each stack, so we should declare this field as an
	// ArrayList<E>.
	private ArrayList<E> items;


	public Stack()
	{
		// Notice that I've specified a type parameter in the creation of the
		// ArrayList<E> object, just as I did in the declaration of the field
		// above.  This is necessary; everywhere we say "ArrayList", we should
		// follow it with its type parameter, so that the compiler will know
		// everything about our intent.
		items = new ArrayList<E>();
	}
	

	// To properly state our intent, we should require that the push() method
	// takes an E as a parameter.  This will prevent a Stack<String> from
	// having anything but a String pushed into it.
	public void push(E e)
	{
		items.add(e);
	}
	

	public void pop()
	{
		if (isEmpty())
		{
			throw new StackEmptyException();
		}
		
		items.remove(items.size() - 1);
	}
	

	// To be absolutely clear, top() should return an E, so that the compiler
	// will know that a Stack<String> can only have Strings returned from it.
	public E top()
	{
		if (isEmpty())
		{
			throw new StackEmptyException();
		}
		
		return items.get(items.size() - 1);
	}
	
	
	public boolean isEmpty()
	{
		return items.size() == 0;
	}


	// This is the method that is required by the Iterable<E> interface.  Its
	// job is to create and return a new iterator.  By design, an iterator
	// over our stack will return the top stack element first and the bottom
	// element last.
	//
	// Note that ArrayList<E> is also iterable, meaning that we could just
	// create and return an iterator over the ArrayList, like this:
	//
	//     return items.iterator();
	//
	// However, this isn't going to support our design.  An iterator over an
	// ArrayList returns the elements starting at cell 0.  In the context of
	// our stack implementation, that won't be right, since it would return
	// the bottom stack element first.
	public Iterator<E> iterator()
	{
		return new StackIterator();
	}
	

	// The StackIterator is an implementation of Java's standard Iterator<E>
	// interface.  An Iterator<E> is an object that can be asked for each
	// successive element in some collection, where each element is an E.
	//
	// An Iterator<E> must support the following methods:
	//
	//    public boolean hasNext()
	//        Returns true if there are more elements to be iterated
	//
	//    public E next()
	//        Returns the next element to be iterated, if any.  If there are
	//        no more elements that haven't been seen, a NoSuchElementException
	//        is thrown.
	//
	//    public void remove()
	//        Removes the last element returned by next().  It is not a
	//        requirement that remove() is supported; if it's not, the method
	//        should throw an UnsupportedOperationException.  (Since we want to
	//        disallow the removal of stack elements other than the top, we'll
	//        have our StackIterator just throw the exception.)
	private class StackIterator implements Iterator<E>
	{
		private int index;


		public StackIterator()
		{
			index = items.size() - 1;
		}
		
		
		public boolean hasNext()
		{
			return index >= 0;
		}
		
		
		public E next()
		{
			if (!hasNext())
			{
				throw new NoSuchElementException();
			}
			
			E item = items.get(index);
			index--;
			return item;
		}


		public void remove()
		{
			// We don't want remove() to be supported, because you're not
			// supposed to remove elements from the middle of a stack!
			throw new UnsupportedOperationException();
		}
	}
}
